## The MISP plugin lists two optional dependencies via `extras_requires`, so
## users can either install `threatbus-misp[zmq]` or `threatbus-misp[kafka]`.
## The plugin needs at least one of these to work properly, so we we need to log
## an error if none of these is installed.
dep_kafka = False
dep_zmq = False
try:
    from confluent_kafka import Consumer

    dep_kafka = True
except ModuleNotFoundError:
    pass
from confuse import Subview
from datetime import datetime
from itertools import product
import json
from multiprocessing import JoinableQueue
import pymisp
from queue import Empty
import threading
import threatbus
from threatbus.data import MessageType, SnapshotEnvelope, SnapshotRequest
from threatbus_misp.message_mapping import (
    attribute_to_stix2_indicator,
    stix2_sighting_to_misp,
    is_whitelisted,
)
from typing import Callable, List, Dict
import warnings

try:
    import zmq

    dep_zmq = True
except ModuleNotFoundError:
    pass


warnings.simplefilter("ignore")  # pymisp produces urllib warnings

"""MISP application plugin for Threat Bus"""


plugin_name: str = "misp"
misp: pymisp.api.PyMISP = None
lock: threading.Lock = threading.Lock()
# filter_config is required for message mapping, but not available when Threat Bus invokes `snapshot()` -> global, initialized on startup
filter_config: List[Dict] = None
workers: List[threatbus.StoppableWorker] = list()


class SightingsPublisher(threatbus.StoppableWorker):
    """
    Reports / publishes true-positive sightings of intelligence items back to the given MISP endpoint.
    """

    def __init__(self, outq: JoinableQueue):
        """
        @param outq The queue from which to forward messages to MISP
        """
        super(SightingsPublisher, self).__init__()
        self.outq = outq

    def run(self):
        global logger, misp, lock
        if not misp:
            return
        while self._running():
            try:
                sighting = self.outq.get(block=True, timeout=1)
            except Empty:
                continue
            misp_sighting = stix2_sighting_to_misp(sighting)
            lock.acquire()
            resp = misp.add_sighting(misp_sighting)
            if not resp or type(resp) is dict and resp.get("message", None):
                logger.error(
                    f"Failed to add sighting to MISP: '{sighting}' Error: {resp}"
                )
            else:
                logger.debug(f"Reported sighting: {resp}")
            lock.release()
            self.outq.task_done()


class KafkaReceiver(threatbus.StoppableWorker):
    """
    Binds a Kafka consumer to the the given host/port. Forwards all received messages to the inq.
    """

    def __init__(self, kafka_config: Subview, inq: JoinableQueue):
        """
        @param kafka_config A configuration object for Kafka binding
        @param inq The queue to which STIX-2 indicators are forwarded to
        """
        super(KafkaReceiver, self).__init__()
        self.kafka_config = kafka_config
        self.inq = inq

    def run(self):
        consumer = Consumer(self.kafka_config["config"].get(dict))
        consumer.subscribe(self.kafka_config["topics"].get(list))
        global logger, filter_config
        while self._running():
            message = consumer.poll(
                timeout=self.kafka_config["poll_interval"].get(float)
            )
            if message is None:
                continue
            if message.error():
                logger.error(f"Kafka error: {message.error()}")
                continue
            try:
                msg = json.loads(message.value())
            except Exception as e:
                logger.error(f"Error decoding Kafka message: {e}")
                continue
            if not is_whitelisted(msg, filter_config):
                continue
            ioc = None
            attr = msg.get("Attribute", None)
            try:
                ioc = attribute_to_stix2_indicator(
                    attr, msg.get("action", None), logger
                )
            except Exception as e:
                logger.warn(f"Failed to parse MISP attribute {attr}: {e}")
            if not ioc:
                # the mapping function returns None e.g., in case a new MISP
                # attribute is added without the `to_ids` flag enabled.
                continue
            self.inq.put(ioc)


class ZmqReceiver(threatbus.StoppableWorker):
    """
    Binds a ZMQ poller to the the given host/port. Forwards all received messages to the inq.
    """

    def __init__(self, zmq_config: Subview, inq: JoinableQueue):
        """
        @param zmq_config A configuration object for ZeroMQ binding
        @param inq The queue to which STIX-2 Indicators are forwarded to
        """
        super(ZmqReceiver, self).__init__()
        self.inq = inq
        self.zmq_config = zmq_config

    def run(self):
        global logger, filter_config
        socket = zmq.Context().socket(zmq.SUB)
        socket.connect(f"tcp://{self.zmq_config['host']}:{self.zmq_config['port']}")
        # TODO: allow reception of more topics, i.e. handle events.
        socket.setsockopt(zmq.SUBSCRIBE, b"misp_json_attribute")
        poller = zmq.Poller()
        poller.register(socket, zmq.POLLIN)

        while self._running():
            socks = dict(poller.poll(timeout=1000))
            if socket not in socks or socks[socket] != zmq.POLLIN:
                continue
            raw = socket.recv()
            _, message = raw.decode("utf-8").split(" ", 1)
            try:
                msg = json.loads(message)
            except Exception as e:
                logger.error(f"Error decoding message {message}: {e}")
                continue
            if not is_whitelisted(msg, filter_config):
                continue
            ioc = None
            attr = msg.get("Attribute", None)
            try:
                ioc = attribute_to_stix2_indicator(
                    attr, msg.get("action", None), logger
                )
            except Exception as e:
                logger.warn(f"Failed to parse MISP attribute {attr}: {e}")
            if not ioc:
                # the mapping function returns None e.g., in case a new MISP
                # attribute is added without the `to_ids` flag enabled.
                continue
            self.inq.put(ioc)


def validate_config(config: Subview):
    assert config, "config must not be None"
    # redact fallback values to allow for omitting configuration blocks
    config["api"].add({})
    config["zmq"].add({})
    config["kafka"].add({})
    config["filter"].add([])

    if config["zmq"].get(dict) and config["kafka"].get(dict):
        raise AssertionError("either use ZeroMQ or Kafka, but not both")

    if type(config["filter"].get()) is not list:
        raise AssertionError("filter must be specified as list")

    if config["api"].get(dict):
        config["api"]["host"].get(str)
        config["api"]["ssl"].get(bool)
        config["api"]["key"].get(str)
    if config["zmq"].get(dict):
        assert (
            dep_zmq
        ), "MISP attribute export is configured via ZeroMQ, but the dependency is not installed. Install `threatbus-misp[zmq]` to use this setting."
        config["zmq"]["host"].get(str)
        config["zmq"]["port"].get(int)
    if config["kafka"].get(dict):
        assert (
            dep_kafka
        ), "MISP attribute export is configured via Apache Kafka, but the dependency is not installed. Install `threatbus-misp[kafka]` to use this setting."
        config["kafka"]["topics"].get(list)
        config["kafka"]["poll_interval"].add(1.0)
        config["kafka"]["poll_interval"].get(float)
        config["kafka"]["config"].get(dict)


@threatbus.app
def snapshot(snapshot_request: SnapshotRequest, result_q: JoinableQueue):
    global logger, misp, lock, filter_config
    if snapshot_request.snapshot_type != MessageType.INDICATOR:
        logger.debug("Sighting snapshot feature not yet implemented.")
        return  # TODO sighting snapshot not yet implemented
    if not misp:
        logger.debug("Cannot perform snapshot request. No MISP API connection.")
        return

    logger.info(f"Executing intel snapshot for time delta {snapshot_request.snapshot}")
    if not filter_config:
        filter_config = [{}]  # this empty whitelist results in a global query

    # build queries for everything that is whitelisted
    for fil in filter_config:
        orgs = fil.get("orgs", [None])
        types = fil.get("types", [None])
        tags_query = misp.build_complex_query(or_parameters=fil.get("tags", []))
        if not tags_query:
            tags_query = None  # explicit None value

        # By API design, orgs and types must be queried value-by-value
        # None-values mean that all values are accepted
        # https://pymisp.readthedocs.io/en/latest/_modules/pymisp/api.html#PyMISP.search
        for (org, type_) in product(orgs, types):
            lock.acquire()
            data = misp.search(
                org=org,
                type_attribute=type_,
                tags=tags_query,
                controller="attributes",
                to_ids=True,
                date_from=datetime.now() - snapshot_request.snapshot,
            )
            lock.release()
            if not data:
                continue
            for attr in data["Attribute"]:
                try:
                    ioc = attribute_to_stix2_indicator(attr, "add", logger)
                except Exception as e:
                    logger.warn(f"Failed to parse MISP attribute {attr}: {e}")
                if ioc:
                    result_q.put(
                        SnapshotEnvelope(
                            snapshot_request.snapshot_type,
                            snapshot_request.snapshot_id,
                            ioc,
                        )
                    )


@threatbus.app
def run(
    config: Subview,
    logging: Subview,
    inq: JoinableQueue,
    subscribe_callback: Callable,
    unsubscribe_callback: Callable,
):
    global logger, filter_config, workers
    logger = threatbus.logger.setup(logging, __name__)
    config = config[plugin_name]
    try:
        validate_config(config)
    except Exception as e:
        logger.fatal("Invalid config for plugin {}: {}".format(plugin_name, str(e)))

    filter_config = config["filter"].get(list)

    # start Attribute-update receiver
    if config["zmq"].get():
        workers.append(ZmqReceiver(config["zmq"], inq))
    elif config["kafka"].get():
        workers.append(KafkaReceiver(config["kafka"], inq))

    # bind to MISP
    if config["api"].get(dict):
        # TODO: MISP instances shall subscribe themselves to threatbus and each
        # subscription shall have an individual outq and receiving thread for intel
        # updates.
        host, key, ssl = (
            config["api"]["host"].get(),
            config["api"]["key"].get(),
            config["api"]["ssl"].get(),
        )
        try:
            global misp, lock
            lock.acquire()
            misp = pymisp.ExpandedPyMISP(url=host, key=key, ssl=ssl)
            lock.release()
        except Exception:
            # TODO: log individual error per MISP subscriber
            logger.error(f"Cannot subscribe to MISP at {host}, using SSL: {ssl}")
            lock.release()
        if not misp:
            logger.error("Failed to start MISP plugin")
            return
    else:
        logger.warning(
            "Starting MISP plugin without API connection, cannot report back sightings or request snapshots."
        )

    outq = JoinableQueue()
    subscribe_callback("stix2/sighting", outq)
    workers.append(SightingsPublisher(outq))
    for w in workers:
        w.start()
    logger.info("MISP plugin started")


@threatbus.app
def stop():
    global logger, workers
    for w in workers:
        if not w.is_alive():
            continue
        w.stop()
        w.join()
    logger.info("MISP plugin stopped")



from datetime import datetime
from functools import partial
from threatbus.data import Operation, ThreatBusSTIX2Constants
import pymisp
from stix2 import (
    AndObservationExpression,
    AutonomousSystem,
    DomainName,
    EmailAddress,
    EqualityComparisonExpression,
    Indicator,
    IPv4Address,
    MACAddress,
    ObjectPath,
    ObservationExpression,
    Sighting,
    URL,
    X509Certificate,
)
from typing import Callable, Dict, List, Union


def observable_value_to_expr(stix2_type, value: str) -> EqualityComparisonExpression:
    """
    Most STIX-2 observables have a 'value' property. This function takes a
    STIX-2 observable type (e.g., DomainName) that uses a value property and
    invokes it's constructor for the given value. Returns a STIX-2
    EqualityComparisonExpression.
    """
    stix_obj = stix2_type(value=value)
    return EqualityComparisonExpression(
        ObjectPath(stix_obj.type, ["value"]), stix_obj.value
    )


def observable_AS_to_expr(number: str) -> EqualityComparisonExpression:
    """
    Autonomous Systems (AS) in STIX-2 have a 'number' property (as opposed to a
    'value'). This function returns a STIX-2 EqualityComparisonExpression for
    an AS number.
    """
    stix_obj = AutonomousSystem(number=int(number))
    return EqualityComparisonExpression(
        ObjectPath(stix_obj.type, ["number"]), stix_obj.number
    )


def observable_x509_to_expr(
    stix2_type, hash_type: str, hash_str: str
) -> EqualityComparisonExpression:
    """
    x509 certificates in STIX-2 have no 'value' property, but an optional field
    'hashes'. This function creates a valid STIX-2 X509Certificate and wraps it
    in an EqualityComparisonExpression.
    """
    hashes = {hash_type: hash_str}  # e.g., {"SHA-256": "aec078..." }
    cert = X509Certificate(hashes=hashes)  # throws an error for invalid hashes
    hash_type = list(cert.hashes.keys())[0]  # STIX-2 corrects case and spelling
    return EqualityComparisonExpression(
        ObjectPath(cert.type, ["hashes", hash_type]), cert.hashes[hash_type]
    )


# Map MISP attribute types [1] [2] to STIX-2 cyber observable objects [3]
# [1]: https://www.misp-standard.org/rfc/misp-standard-core.html#type
# [2]: https://www.circl.lu/doc/misp/categories-and-types/#types
# [3]: https://docs.oasis-open.org/cti/stix/v2.1/cs01/stix-v2.1-cs01.html#_mlbmudhl16lr

attribute_type_map: Dict[
    str, Union[Callable[[str], Union[EqualityComparisonExpression, str]], None]
] = {
    # TODO: limited to MISP's `network activity` observables for now.
    # Discuss & expand this if needed
    "ip": partial(
        observable_value_to_expr, IPv4Address
    ),  # https://stix2.readthedocs.io/en/latest/api/stix2.v21.html?highlight=IPv4Address#stix2.v21.IPv4Address
    "ip-src": partial(
        observable_value_to_expr, IPv4Address
    ),  # https://stix2.readthedocs.io/en/latest/api/stix2.v21.html?highlight=IPv4Address#stix2.v21.IPv4Address
    "ip-dst": partial(
        observable_value_to_expr, IPv4Address
    ),  # https://stix2.readthedocs.io/en/latest/api/stix2.v21.html?highlight=IPv4Address#stix2.v21.IPv4Address
    # "ip-dst|port" -> compound types are mapped via their individual parts
    # "ip-src|port" -> compound types are mapped via their individual parts
    "port": None,  # TODO no appropriate mapping to a STIX-2 observable
    "hostname": partial(
        observable_value_to_expr, DomainName
    ),  # https://stix2.readthedocs.io/en/latest/api/stix2.v21.html#stix2.v21.DomainName
    "domain": partial(
        observable_value_to_expr, DomainName
    ),  # https://stix2.readthedocs.io/en/latest/api/stix2.v21.html#stix2.v21.DomainName
    # "domain|ip" -> compound types are mapped via their individual parts
    "mac-address": partial(
        observable_value_to_expr, MACAddress
    ),  # https://stix2.readthedocs.io/en/latest/api/stix2.v21.html#stix2.v21.MACAddress
    "mac-eui-64": partial(
        observable_value_to_expr, MACAddress
    ),  # https://stix2.readthedocs.io/en/latest/api/stix2.v21.html#stix2.v21.MACAddress
    "email": partial(
        observable_value_to_expr, EmailAddress
    ),  # https://docs.oasis-open.org/cti/stix/v2.1/cs01/stix-v2.1-cs01.html#_wmenahkvqmgj
    "email-dst": partial(
        observable_value_to_expr, EmailAddress
    ),  # https://docs.oasis-open.org/cti/stix/v2.1/cs01/stix-v2.1-cs01.html#_wmenahkvqmgj
    "email-src": partial(
        observable_value_to_expr, EmailAddress
    ),  # https://docs.oasis-open.org/cti/stix/v2.1/cs01/stix-v2.1-cs01.html#_wmenahkvqmgj
    "eppn": partial(
        observable_value_to_expr, EmailAddress
    ),  # EduPersonPricincipalName, looks like an Email. https://github.com/MISP/MISP/issues/5448 https://docs.oasis-open.org/cti/stix/v2.1/cs01/stix-v2.1-cs01.html#_wmenahkvqmgj
    "url": partial(
        observable_value_to_expr, URL
    ),  # https://docs.oasis-open.org/cti/stix/v2.1/cs01/stix-v2.1-cs01.html#_ah3hict2dez0
    "uri": partial(
        observable_value_to_expr, URL
    ),  # https://docs.oasis-open.org/cti/stix/v2.1/cs01/stix-v2.1-cs01.html#_ah3hict2dez0
    "user-agent": None,  # no appropriate mapping to a STIX-2 observable
    "http-method": None,  # no appropriate mapping to a STIX-2 observable
    "AS": observable_AS_to_expr,  # https://docs.oasis-open.org/cti/stix/v2.1/cs01/stix-v2.1-cs01.html#_27gux0aol9e3
    "snort": None,  # no appropriate mapping to a STIX-2 observable
    "pattern-in-file": None,  # no appropriate mapping to a STIX-2 observable
    "filename-pattern": None,  # no appropriate mapping to a STIX-2 observable
    "stix2-pattern": lambda val: val[1:-1]
    if val.startswith("[") and val.endswith("]")
    else val,  # remove brackets, if any
    "pattern-in-traffic": None,  # no appropriate mapping to a STIX-2 observable
    "attachment": None,  # no appropriate mapping to a STIX-2 observable
    "comment": None,  # no appropriate mapping to a STIX-2 observable
    "text": None,  # no appropriate mapping to a STIX-2 observable
    "x509-fingerprint-md5": partial(
        observable_x509_to_expr, X509Certificate, "md5"
    ),  # https://docs.oasis-open.org/cti/stix/v2.1/cs01/stix-v2.1-cs01.html#_8abcy1o5x9w1
    "x509-fingerprint-sha1": partial(
        observable_x509_to_expr, X509Certificate, "sha1"
    ),  # https://docs.oasis-open.org/cti/stix/v2.1/cs01/stix-v2.1-cs01.html#_8abcy1o5x9w1
    "x509-fingerprint-sha256": partial(
        observable_x509_to_expr, X509Certificate, "sha256"
    ),  # https://docs.oasis-open.org/cti/stix/v2.1/cs01/stix-v2.1-cs01.html#_8abcy1o5x9w1
    "ja3-fingerprint-md5": None,  # no appropriate mapping to a STIX-2 observable
    "jarm-fingerprint": None,  # no appropriate mapping to a STIX-2 observable
    "hassh-md5": None,  # no appropriate mapping to a STIX-2 observable
    "hasshserver-md5": None,  # no appropriate mapping to a STIX-2 observable
    "other": None,  # no appropriate mapping to a STIX-2 observable
    "hex": None,  # no appropriate mapping to a STIX-2 observable
    "cookie": None,  # no appropriate mapping to a STIX-2 observable
    # "hostname|port" -> compound types are mapped via their individual parts
    "bro": None,  # no appropriate mapping to a STIX-2 observable
    "zeek": None,  # no appropriate mapping to a STIX-2 observable
    # "anonymised" -> compound types are mapped via their individual parts
    "community-id": None,  # no appropriate mapping to a STIX-2 observable
    "email-subject": None,  # no appropriate mapping to a STIX-2 observable
    "favicon-mmh3": None,  # no appropriate mapping to a STIX-2 observable
}


def get_tags(misp_attr: dict):
    """
    Tags are attached as list of objects to MISP attributes. Returns a list of
    all tag names.
    @param The MISP attribute to get the tag names from
    @return A list of tag names
    """
    return [
        t.get("name", None)
        for t in misp_attr.get("Tag", [])
        if t and t.get("name", None)
    ]


def is_whitelisted(misp_msg: dict, filter_config: List[Dict]):
    """
    Compares the given MISP message against every filter in the filter_config
    list. Returns True if the filter_config is empty or if the event is
    whitelisted according to at least one filter from the filter_config.
    @param misp_msg The MISP message to check
    @param filter_config A list of whitelist-filters for "orgs", "tags", and
        "types"
    @return Boolean flag to indicate if the event is whitelisted
    """
    if not misp_msg:
        return False
    event = misp_msg.get("Event", None)
    attr = misp_msg.get("Attribute", None)
    if not event or not attr:
        return False
    org_id = event.get("org_id", None)
    intel_type = attr.get("type", None)
    tags = get_tags(attr)
    if not org_id or not intel_type:
        return False
    if not filter_config:
        # no whitelist = allow all
        return True
    for fil in filter_config:
        if (
            (not fil.get("orgs", None) or org_id in fil["orgs"])
            and (not fil.get("types", None) or intel_type in fil["types"])
            and (
                not fil.get("tags", None)
                or len(set(tags).intersection(set(fil["tags"]))) > 0
            )
        ):
            return True
    return False


def stix2_indicator_id(attr_uuid: str) -> str:
    """
    Converts any given MISP attribute UUID to a STIX-2 indicator ID.
    @param attr_id The MISP attribute ID to convert
    @return a valid STIX-2 Indicator ID
    """
    return f"indicator--{attr_uuid}"


def misp_id(stix2_indicator_uuid: str) -> str:
    """
    Converts any given STIX-2 indicator ID to a valid UUID.
    @param stix2_indicator_uuid The STIX-2 Indicator ID to convert
    @return a valid uuid
    """
    return stix2_indicator_uuid[len("indicator--") :]


def to_equality_expr(
    misp_attr_type: str, misp_value: str
) -> Union[EqualityComparisonExpression, None]:
    """
    Returns a valid STIX-2 EqualityComparisonExpression for a given MISP
    attribute type and a point-ioc value.
    @param misp_attr_type The MISP attribute type (e.g., 'domain')
    @param misp_value A MISP point-ioc value (e.g., 'evil.com')
    @return EqualityComparisonExpression
    """
    stix2_create_func = attribute_type_map.get(misp_attr_type, None)
    if not stix2_create_func:
        raise ValueError(
            f"Cannot find STIX-2 mapping for MISP attribute type '{misp_attr_type}'"
        )
    return stix2_create_func(misp_value)


def attribute_to_stix2_indicator(
    misp_attribute: dict, action: str, logger
) -> Union[None, Indicator]:
    """
    Maps the given MISP attribute to a STIX-2 cyber observable object.
    @param misp_attribute The MISP attribute to map to STIX-2
    @param action A string from MISP, describing the action for the attribute
        (either 'add', 'edit', or 'delete')
    @param logger The logger instance from the calling function
    @return The mapped STIX-2 Indicator or None
    @throws ValueError for invalid inputs
    """
    # parse MISP attribute
    if not misp_attribute or not action or action not in ["add", "edit", "delete"]:
        raise ValueError(
            f"Cannot parse MISP attribute {misp_attribute} with action '{action}'."
        )
    ## parse MISP values
    to_ids = misp_attribute.get("to_ids", None)
    misp_ts = misp_attribute.get("timestamp", None)
    misp_id = misp_attribute.get("uuid", None)
    misp_attr_type = misp_attribute.get("type", None)
    misp_attr_value = misp_attribute.get("value", None)
    if (
        to_ids is None
        or not misp_ts
        or not misp_id
        or not misp_attr_type
        or not misp_attr_value
    ):
        raise ValueError(
            "Incomplete MISP attribute, missing one of the following properties: `timestamp`, `to_ids`, `uuid`, `type`, `value`"
        )
    if to_ids == False and action != "edit":
        # A new attribute was created, but the `to_ids` flag is toggled off.
        return None

    custom_properties = {}
    if action == "edit" and to_ids == False:
        # A previously published MISP attribute was edited and the `to_ids` flag
        # was toggled off. We mark this with a custom STIX-2 property as deleted
        custom_properties[
            ThreatBusSTIX2Constants.X_THREATBUS_UPDATE.value
        ] = Operation.REMOVE.value

    stix2_timestamp = datetime.fromtimestamp(int(misp_ts))
    stix2_id = stix2_indicator_id(misp_id)

    ## parse compound MISP intel:
    if "|" in misp_attr_type:
        obs_exprs = []  # list to hold all ObservationExpressions
        value_splits = misp_attr_value.split("|")
        for idx, misp_attr_type_split in enumerate(misp_attr_type.split("|")):
            eq_expr = to_equality_expr(misp_attr_type_split, value_splits[idx])
            obs_expr = ObservationExpression(eq_expr)
            obs_exprs.append(obs_expr)
        return Indicator(
            id=stix2_id,
            pattern_type="stix",
            pattern=AndObservationExpression(obs_exprs),
            created=stix2_timestamp,
            custom_properties=custom_properties,
        )

    ## parse point-IoC
    eq_expr = to_equality_expr(misp_attr_type, misp_attr_value)
    return Indicator(
        id=stix2_id,
        pattern_type="stix",
        pattern=ObservationExpression(eq_expr),
        created=stix2_timestamp,
        custom_properties=custom_properties,
    )


def stix2_sighting_to_misp(sighting: Sighting):
    """
    Maps the STIX-2 sighting format to a MISP sighting.
    @param sighting A STIX-2 Sighting object
    @return the mapped MISP sighting object or None
    """
    if not sighting or type(sighting) != Sighting:
        return None

    misp_sighting = pymisp.MISPSighting()
    source = None
    if "x_threatbus_source" in sighting.object_properties():
        source = str(sighting.x_threatbus_source)
    misp_sighting.from_dict(
        id=misp_id(sighting.sighting_of_ref),
        source=source,
        type="0",  # true positive sighting: https://www.misp-standard.org/rfc/misp-standard-core.html#sighting
        timestamp=sighting.created,
    )
    return misp_sighting



from datetime import datetime
import json
from pymisp import MISPSighting
from stix2 import Indicator, Sighting
from stix2.exceptions import InvalidValueError
from threatbus.data import Operation
from threatbus_misp.message_mapping import (
    attribute_type_map,
    attribute_to_stix2_indicator,
    stix2_sighting_to_misp,
    is_whitelisted,
    get_tags,
)
import unittest


class TestMessageMapping(unittest.TestCase):
    def setUp(self):
        self.raw_ts = 1579104545
        self.ts = datetime.fromtimestamp(self.raw_ts)
        self.domain_ioc = "example.com"
        self.ip_ioc = "example.com"
        self.pattern = f"[domain-name:value = '{self.domain_ioc}'] AND [ipv4-addr:value = '{self.ip_ioc}']"
        self.attr_uuid = "5e1f2787-fcfc-4718-a58a-00b4c0a82f06"
        self.indicator_id = f"indicator--{self.attr_uuid}"
        self.indicator = Indicator(
            id=self.indicator_id,
            created=self.ts,
            pattern_type="stix",
            pattern=self.pattern,
        )
        self.sighting = Sighting(created=self.ts, sighting_of_ref=self.indicator_id)
        self.valid_misp_attribute = {
            "id": self.id,
            "event_id": "1",
            "object_id": "0",
            "object_relation": None,
            "category": "Network activity",
            "type": "domain|ip",
            "value1": self.domain_ioc,
            "value2": self.ip_ioc,
            "to_ids": True,
            "uuid": self.attr_uuid,
            "timestamp": str(self.raw_ts),
            "distribution": "5",
            "sharing_group_id": "0",
            "comment": "",
            "deleted": False,
            "disable_correlation": False,
            "value": f"{self.domain_ioc}|{self.ip_ioc}",
            "Sighting": [],
            "Tag": [],
        }
        self.valid_misp_msg = json.loads(
            """{
        "Attribute": {
            "id": "1",
            "event_id": "1",
            "object_id": "0",
            "object_relation": null,
            "category": "Network activity",
            "type": "ip-src",
            "value1": "6.6.6.6",
            "value2": "",
            "to_ids": true,
            "uuid": "5f7b2284-bdd8-4ef4-b457-032ac0a82f02",
            "timestamp": "1601906335",
            "distribution": "5",
            "sharing_group_id": "0",
            "comment": "",
            "deleted": false,
            "disable_correlation": false,
            "first_seen": null,
            "last_seen": null,
            "value": "6.6.6.6",
            "Tag": [
            {
                "id": "3",
                "name": "baz",
                "colour": "#05ff00",
                "exportable": true
            },
            {
                "id": "1",
                "name": "foo",
                "colour": "#ff0000",
                "exportable": true
            }
            ],
            "Sighting": []
        },
        "Event": {
            "id": "1",
            "date": "2020-10-05",
            "info": "evil",
            "uuid": "5f7b2277-1ebc-4101-9152-032ac0a82f02",
            "published": false,
            "analysis": "0",
            "threat_level_id": "1",
            "org_id": "1",
            "orgc_id": "1",
            "distribution": "1",
            "sharing_group_id": "0",
            "Orgc": {
            "id": "1",
            "uuid": "5f7b21ea-52a4-49f0-87df-032ac0a82f02",
            "name": "ORGNAME"
            }
        },
        "action": "edit"
        }
        """
        )

    def test_invalid_stix_sightings(self):
        self.assertIsNone(stix2_sighting_to_misp(None))
        self.assertIsNone(stix2_sighting_to_misp("Hello"))
        self.assertIsNone(stix2_sighting_to_misp(self))

    def test_invalid_misp_attributes(self):
        self.assertRaises(ValueError, attribute_to_stix2_indicator, None, None, None)
        self.assertRaises(
            ValueError, attribute_to_stix2_indicator, None, "delete", None
        )
        self.assertRaises(
            ValueError,
            attribute_to_stix2_indicator,
            self.valid_misp_attribute,
            None,
            None,
        )

    def test_invalid_action(self):
        self.assertRaises(
            ValueError,
            attribute_to_stix2_indicator,
            self.valid_misp_attribute,
            "INVALID_ACTION",
            None,
        )

    def test_valid_misp_attributes(self):
        indicator = attribute_to_stix2_indicator(self.valid_misp_attribute, "add", None)
        self.assertEqual(indicator.type, self.indicator.type)
        self.assertEqual(indicator.id, self.indicator.id)
        self.assertEqual(indicator.created, self.indicator.created)
        self.assertEqual(indicator.pattern, self.indicator.pattern)

        single_value_valid_misp_attribute = self.valid_misp_attribute.copy()
        single_value_valid_misp_attribute["type"] = "domain"
        single_value_valid_misp_attribute["value"] = self.domain_ioc
        single_value_valid_misp_attribute["value1"] = self.domain_ioc
        single_value_valid_misp_attribute["value2"] = None
        indicator = attribute_to_stix2_indicator(
            single_value_valid_misp_attribute, "add", None
        )
        self.assertEqual(indicator.type, self.indicator.type)
        self.assertEqual(indicator.id, self.indicator.id)
        self.assertEqual(indicator.created, self.ts)
        self.assertEqual(
            indicator.pattern, f"[domain-name:value = '{self.domain_ioc}']"
        )
        # test that no custom properties are set
        for prop in indicator.object_properties():
            self.assertTrue(not prop.startswith("x_threatbus_"))

    def test_valid_misp_attribute_removal(self):
        valid_misp_attribute = self.valid_misp_attribute.copy()
        valid_misp_attribute["to_ids"] = False
        indicator = attribute_to_stix2_indicator(valid_misp_attribute, "edit", None)
        self.assertEqual(indicator.x_threatbus_update, Operation.REMOVE.value)

    def test_valid_stix_sighting(self):
        misp_sighting = stix2_sighting_to_misp(self.sighting)
        self.assertIsNotNone(misp_sighting)
        self.assertEqual(type(misp_sighting), MISPSighting)

    def test_invalid_tags(self):
        attr = self.valid_misp_msg["Attribute"]
        del attr["Tag"]
        tags = get_tags(attr)
        self.assertEqual(tags, [])

        attr["Tag"] = []
        tags = get_tags(attr)
        self.assertEqual(tags, [])

        attr["Tag"] = [{"some": "property"}]
        tags = get_tags(attr)
        self.assertEqual(tags, [])

    def test_valid_tags(self):
        attr = self.valid_misp_msg["Attribute"]
        tags = get_tags(attr)
        self.assertEqual(sorted(tags), sorted(["foo", "baz"]))

    def test_invalid_message_is_whitelisted(self):
        event = self.valid_misp_msg["Event"]
        filter_config = {}  # nothing whitelisted means everything is whitelisted

        del self.valid_misp_msg["Event"]
        self.assertFalse(is_whitelisted(self.valid_misp_msg, filter_config))

        self.valid_misp_msg["Event"] = event
        del self.valid_misp_msg["Attribute"]
        self.assertFalse(is_whitelisted(self.valid_misp_msg, filter_config))

    def test_valid_message_is_whitelisted(self):

        # positive tests with one filter:
        filter_config = []
        self.assertTrue(is_whitelisted(self.valid_misp_msg, filter_config))

        filter_config = [{"orgs": ["1", "100"]}]
        self.assertTrue(is_whitelisted(self.valid_misp_msg, filter_config))

        filter_config = [{"tags": ["foo", "ASDF"]}]
        self.assertTrue(is_whitelisted(self.valid_misp_msg, filter_config))

        filter_config = [{"types": ["domain", "ip-src"]}]
        self.assertTrue(is_whitelisted(self.valid_misp_msg, filter_config))

        filter_config = [{"orgs": ["1"], "tags": ["foo"]}]
        self.assertTrue(is_whitelisted(self.valid_misp_msg, filter_config))

        filter_config = [{"orgs": ["1"], "types": ["ip-src"]}]
        self.assertTrue(is_whitelisted(self.valid_misp_msg, filter_config))

        filter_config = [{"tags": ["foo"], "types": ["ip-src"]}]
        self.assertTrue(is_whitelisted(self.valid_misp_msg, filter_config))

        filter_config = [{"orgs": ["1"], "tags": ["foo"], "types": ["ip-src"]}]
        self.assertTrue(is_whitelisted(self.valid_misp_msg, filter_config))

        # positive tests with multiple filters where only one matches
        filter_config = [{"orgs": ["1"]}, {"orgs": ["5"]}]
        self.assertTrue(is_whitelisted(self.valid_misp_msg, filter_config))

        filter_config = [{"orgs": ["1"]}, {"types": ["ASDF"]}]
        self.assertTrue(is_whitelisted(self.valid_misp_msg, filter_config))

        filter_config = [{"tags": ["foo"]}, {"types": ["ASDF"]}]
        self.assertTrue(is_whitelisted(self.valid_misp_msg, filter_config))

        # positive tests with multiple filters where many match
        filter_config = [{"tags": ["foo"]}, {"types": ["ip-src"]}]
        self.assertTrue(is_whitelisted(self.valid_misp_msg, filter_config))

        filter_config = [{"tags": ["baz"]}, {"org": ["1"]}]
        self.assertTrue(is_whitelisted(self.valid_misp_msg, filter_config))

        # negative tests with one filter:
        filter_config = [{"orgs": ["100"]}]  # org doesn't match
        self.assertFalse(is_whitelisted(self.valid_misp_msg, filter_config))

        filter_config = [{"tags": ["ASDF"]}]  # tag doesn't match
        self.assertFalse(is_whitelisted(self.valid_misp_msg, filter_config))

        filter_config = [{"types": ["domain"]}]  # type doesn't match
        self.assertFalse(is_whitelisted(self.valid_misp_msg, filter_config))

        filter_config = [{"orgs": ["1"], "tags": ["ASDF"]}]  # tag doesn't match
        self.assertFalse(is_whitelisted(self.valid_misp_msg, filter_config))

        filter_config = [{"types": ["ip-src"], "tags": ["ASDF"]}]  # tag doesn't match
        self.assertFalse(is_whitelisted(self.valid_misp_msg, filter_config))

        filter_config = [{"orgs": ["1"], "types": ["domain"]}]  # type doesn't match
        self.assertFalse(is_whitelisted(self.valid_misp_msg, filter_config))

        filter_config = [{"tags": ["foo"], "types": ["domain"]}]  # type doesn't match
        self.assertFalse(is_whitelisted(self.valid_misp_msg, filter_config))

        filter_config = [{"types": ["src-ip"], "orgs": ["100"]}]  # org doesn't match
        self.assertFalse(is_whitelisted(self.valid_misp_msg, filter_config))

        filter_config = [{"tags": ["foo"], "orgs": ["100"]}]  # org doesn't match
        self.assertFalse(is_whitelisted(self.valid_misp_msg, filter_config))

        # negative tests with multiple filters where none matches
        filter_config = [{"orgs": ["100"]}, {"orgs": ["5"]}]
        self.assertFalse(is_whitelisted(self.valid_misp_msg, filter_config))

        filter_config = [{"orgs": ["100"]}, {"types": ["ASDF"]}]
        self.assertFalse(is_whitelisted(self.valid_misp_msg, filter_config))

        filter_config = [{"tags": ["ASDF"]}, {"types": ["ASDF"]}]
        self.assertFalse(is_whitelisted(self.valid_misp_msg, filter_config))

    def test_attribute_type_map(self):
        def assert_expectation(attr_type, value, expectation):
            if expectation is None:
                self.assertIsNone(attribute_type_map[attr_type])
                return
            create_func = attribute_type_map[attr_type]
            if expectation is InvalidValueError:
                self.assertRaises(expectation, create_func, value)
                return
            self.assertEqual(str(create_func(value)), expectation)

        misp_types_values_expectations = [
            ("ip", "6.6.6.6", "ipv4-addr:value = '6.6.6.6'"),
            ("ip-src", "6.6.6.6", "ipv4-addr:value = '6.6.6.6'"),
            ("ip-dst", "6.6.6.6", "ipv4-addr:value = '6.6.6.6'"),
            ("port", None, None),
            ("hostname", "evil.com", "domain-name:value = 'evil.com'"),
            ("domain", "evil.com", "domain-name:value = 'evil.com'"),
            (
                "mac-address",
                "00:e0:f0:a0:a1:a2",
                "mac-addr:value = '00:e0:f0:a0:a1:a2'",
            ),
            ("mac-eui-64", "00:e0:f0:a0:a1:a2", "mac-addr:value = '00:e0:f0:a0:a1:a2'"),
            ("email", "foo@bar.com", "email-addr:value = 'foo@bar.com'"),
            ("email-dst", "foo@bar.com", "email-addr:value = 'foo@bar.com'"),
            ("email-src", "foo@bar.com", "email-addr:value = 'foo@bar.com'"),
            ("eppn", "foo@bar.com", "email-addr:value = 'foo@bar.com'"),
            (
                "url",
                "https://example.com/foo/bar",
                "url:value = 'https://example.com/foo/bar'",
            ),
            (
                "uri",
                "https://example.com/foo/bar",
                "url:value = 'https://example.com/foo/bar'",
            ),
            ("user-agent", None, None),
            ("http-method", None, None),
            ("AS", "123456", "autonomous-system:number = 123456"),
            ("snort", None, None),
            ("pattern-in-file", None, None),
            ("filename-pattern", None, None),
            (
                "stix2-pattern",
                "ipv4-addr:value = '6.6.6.6'",
                "ipv4-addr:value = '6.6.6.6'",
            ),  # w/o brackets
            (
                "stix2-pattern",
                "[ipv4-addr:value = '6.6.6.6']",
                "ipv4-addr:value = '6.6.6.6'",
            ),  # w/ brackets
            ("pattern-in-traffic", None, None),
            ("attachment", None, None),
            ("comment", None, None),
            ("text", None, None),
            (
                "x509-fingerprint-md5",
                "3d8c1104c6c2482eb4afc4458109a84e",
                "x509-certificate:hashes.MD5 = '3d8c1104c6c2482eb4afc4458109a84e'",
            ),
            ("x509-fingerprint-md5", "fooo", InvalidValueError),  # invalid hash
            (
                "x509-fingerprint-sha1",
                "6e6187d58483457f86eae49315a0a72d7543459a",
                "x509-certificate:hashes.'SHA-1' = '6e6187d58483457f86eae49315a0a72d7543459a'",
            ),
            ("x509-fingerprint-sha1", "fooo", InvalidValueError),  # invalid hash
            (
                "x509-fingerprint-sha256",
                "6aab1f6a6f58fb33fbf15050d7aae38a870e477f42a1707ab3aa243b512cdb6b",
                "x509-certificate:hashes.'SHA-256' = '6aab1f6a6f58fb33fbf15050d7aae38a870e477f42a1707ab3aa243b512cdb6b'",
            ),
            ("x509-fingerprint-sha256", "fooo", InvalidValueError),  # invalid hash
            ("ja3-fingerprint-md5", None, None),
            ("jarm-fingerprint", None, None),
            ("hassh-md5", None, None),
            ("hasshserver-md5", None, None),
            ("other", None, None),
            ("hex", None, None),
            ("cookie", None, None),
            ("bro", None, None),
            ("zeek", None, None),
            ("community-id", None, None),
            ("email-subject", None, None),
            ("favicon-mmh3", None, None),
        ]
        for entry in misp_types_values_expectations:
            assert_expectation(*entry)
